/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is LightSMS.
 *
 * The Initial Developer of the Original Code is
 * Byungwook Kang.
 * Portions created by the Initial Developer are Copyright (C) 2007
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

/***********************************************************
constants
***********************************************************/

const nsILightSMS = Components.interfaces.nsILightSMS;

// UUID uniquely identifying our component
const CLASS_ID = Components.ID("{20fd7fe8-248d-11dc-8314-0800200c9a66}");

// description
const CLASS_NAME = "LightSMS Service";

// textual unique identifier
const CONTRACT_ID = "@mozilla.org/LightSMS;1";





const DATA_VASP_ID=0;
const DATA_VAS_ID=1;
const DATA_USER_KEY=2;
const DATA_USER_LEVEL=3;
const DATA_USER_MDN=4;
const DATA_USER_NUM=5;
const DATA_FREE_COUNT=6;
const DATA_CASH_KEY=7;
const DATA_COUPON_COUNT=8;
const DATA_REFERER=9;

const STAGE_READY=20;
const STAGE_SEND=21;

const CMD_LOGIN=0;
const CMD_SEND=1;

const RETRY_LIMIT=3;



const ST_PRE         = 0;
const ST_PRE_RES     = 1;
const ST_LOGIN       = 10;
const ST_LOGIN_RES   = 11;
const ST_DATA        = 20;
const ST_DATA_RES    = 21;

const LOGIN_URL      = 0;
const LOGIN_ID       = 1;
const LOGIN_PW       = 2;
const LOGIN_DATA     = 3;
const LOGIN_POST     = 1;

const ST_SEND        = 30;
const ST_SEND_RES    = 31;

const M_SEND         = 0;
const M_RECV         = 1;
const M_MSG          = 2;
const M_INDEX        = 3;
const M_TOTAL        = 4;

const formSubmitURL = "user";
const dout=Components.utils.reportError;

/***********************************************************
class definition
***********************************************************/

//class constructor
function LightSMS() {
  var observerService = Components.classes["@mozilla.org/observer-service;1"]
                                .getService(Components.interfaces.nsIObserverService);
  observerService.addObserver(this, "http-on-modify-request", false);
  observerService.addObserver(this, "http-on-examine-response", false);

  var prefService = Components.classes["@mozilla.org/preferences-service;1"]
                              .getService(Components.interfaces.nsIPrefService);
  this.prefBranch = prefService.getBranch("extensions.lightsms.");
  this.prefBranch.QueryInterface(Components.interfaces.nsIPrefBranch2);
  this.prefBranch.addObserver("", this, false);

  if ("@mozilla.org/login-manager;1" in Components.classes){
    this.loginManager = Components.classes["@mozilla.org/login-manager;1"]
                           .getService(Components.interfaces.nsILoginManager);
  }else if ("@mozilla.org/passwordmanager;1" in Components.classes){
    this.passwordManager = Components.classes["@mozilla.org/passwordmanager;1"]
                                  .getService(Components.interfaces.nsIPasswordManager);
  }

  var ios = Components.classes["@mozilla.org/network/io-service;1"]
                    .getService(Components.interfaces.nsIIOService);
  var fileHandler = ios.getProtocolHandler("file")
                       .QueryInterface(Components.interfaces.nsIFileProtocolHandler);
  var loader =  Components.classes["@mozilla.org/moz/jssubscript-loader;1"]
                  .getService(Components.interfaces.mozIJSSubScriptLoader);
  var file;
  file=__LOCATION__.parent;
  var dir=fileHandler.getURLSpecFromFile(file);
  loader.loadSubScript(dir+"cookie.js");
  file.append("scripts");
  this.dir0=fileHandler.getURLSpecFromFile(file);
  try{
    loader.loadSubScript(this.dir0+"order.dat",this);
  }catch(e){dout(e);}

  this.scripts={};
  if(file.exists()){
    var entries = file.directoryEntries;
    while(entries.hasMoreElements())
    {
      var t = entries.getNext();
      t.QueryInterface(Components.interfaces.nsIFile);
      var s=t.leafName;
      var fnd;
      if(fnd=s.match(/((lib-)?.+).js$/)){
        if(fnd[2]){//load library
          try{
            loader.loadSubScript(this.dir0+fnd[0]);
          }catch(e){
dout(fnd[0]+" "+e);
          }
        }else{
          this.scripts[fnd[1]]=null;
        }
      }
    }
  }  
}

// class definition
LightSMS.prototype = {
  listeners : new Array(),
  inited : false,
  initCalled : false,
  
  /*
  userData:[],
  stage:0,
  onreq:false,
  cmd:CMD_LOGIN,
  recv:null,
  msg:null,
  retry:0,
  */
  
  addListener: function(aListener) {
    this.listeners.push(aListener);
    if(!this.inited){
      if(!this.initCalled){
        this.initCalled=true;
        this.init();
        /*for each(var l in this.listeners){
          l.onStateChange(nsIWebMailNotifier.ST_INIT,0,null);
          for each(var o in this.webMailHandlers){
            if(o.enabled){
              l.onStateChange(o.ind,o.calcCount(),encodeObject(o.mailData));
            }
          }
        }*/
        this.inited=true;
      }
    }else{
      if(this.worker&&this.worker.count>=0){
        aListener.onStateChange(nsILightSMS.ST_LOGIN,nsILightSMS.RS_SUCC,null);
        aListener.onStateChange(nsILightSMS.ST_FREECOUNT,this.worker.count,this.getWorkerStr());
      }
    }
  },
  init: function(){
    this.msg=[];
    this.bundle = Components.classes["@mozilla.org/intl/stringbundle;1"]
             .getService(Components.interfaces.nsIStringBundleService)
             .createBundle("chrome://lightsms/locale/lightsms.properties"),  
    this.initHandlers(true);
    this.checkFirstrun();    
  },  
  checkFirstrun:function(){
    var ver = -1;
    var gExtensionManager = Components.classes["@mozilla.org/extensions/manager;1"]
                            .getService(Components.interfaces.nsIExtensionManager);
    var current = gExtensionManager.getItemForID("{0531cfe8-2487-11dc-8314-0800200c9a66}").version;
    try{
      ver = this.prefBranch.getCharPref("version");
    }catch(e){
    }finally{
      if (ver!=current){
        this.prefBranch.setCharPref("version",current);
        var timer = Components.classes["@mozilla.org/timer;1"].createInstance(Components.interfaces.nsITimer);
        var tmp={
          notify:function(aTimer){
            var win=LightSMS.prototype.getWindow("navigator:browser");
            var browser=win.getBrowser();
            var tab=browser.addTab("http://lightsms.mozdev.org/updated.php");
            browser.selectedTab = tab;
          }
        };
        timer.initWithCallback(tmp,1500,Components.interfaces.nsITimer.TYPE_ONE_SHOT); //Firefox 2 fix - or else tab will get closed
      }
    }
  },  
  sendSMS: function(sender,recv,msg){
    recv=recv.split(";");
    var n=0;
    for each(var o in recv){
      if(this.getNumber(o))this.msg.push([sender,o,msg,n++]);
    }
    for each(var o in this.msg)o[M_TOTAL]=n;
 
    if(this.worker==null){
      this.checkAll();
      return;
    }else{
      this.worker.checkData();
    }
  },
  removeListener: function(aListener){
    for(var i=0;i<this.listeners.length;i++){
      if(this.listeners[i]==aListener){
          this.listeners.splice(i,1);
          break;
      }
    }
  },  
  QueryInterface: function(aIID)
  {
    if (!aIID.equals(nsILightSMS) &&
        !aIID.equals(Components.interfaces.nsISupports))
      throw Components.results.NS_ERROR_NO_INTERFACE;
    return this;
  }
};
LightSMS.prototype.initHandler=function(obj,id,user,pass) {
  this.handlers.push(obj);          
  obj.id=id;        
  obj.user=user;     
  obj.password=pass;
  obj.inited=false;
  if(this.order[id])obj.key=this.order[id].toString();  
}
LightSMS.prototype.initHandlers=function(onInit){
  this.worker=null;
  this.stopAll();
  this.handlers=[];
  var str="chrome://lightsms/accounts/";
  if(this.loginManager){
    for(var id in this.scripts){
      var logins = this.loginManager.findLogins({},str+id,formSubmitURL, null);
      for each (var o in logins) {
        var obj=new Handler(this);
        this.initHandler(obj,id,o.username,o.password);
      }
    }
  }else{
    var e = this.passwordManager.enumerator;
    while (e.hasMoreElements()) {
      try {
        var pass = e.getNext().QueryInterface(Components.interfaces.nsIPassword);
        if (pass.host.indexOf(str)==0) {
          var id=pass.host.substring(str.length);
          var obj=new Handler(this);
          this.initHandler(obj,id,pass.user,pass.password);
        }
      }catch(ex) {}
    }
  }
  this.handlers.sort(this.sortFunc);  
  
  var loader =  Components.classes["@mozilla.org/moz/jssubscript-loader;1"]
                  .getService(Components.interfaces.mozIJSSubScriptLoader);
  var i;
  for(i=0;i<this.handlers.length;++i){
    var obj=this.handlers[i];
    try{
      obj.enabled=this.prefBranch.getBoolPref("accounts."+obj.id+"."+obj.user+".enabled");
    }catch(e){}
    try{
      obj.order=this.prefBranch.getIntPref("accounts."+obj.id+"."+obj.user+".order");
    }catch(e){}          
    if(!obj.enabled||obj.inited)continue;
    try{
      this.loadScript(obj.id);
      var scr=this.scripts[obj.id];
      for(var k in scr){
        obj[k]=scr[k];
      }
      if(obj.init)obj.init();
      if(obj.loginData){
        var post=(obj.loginData[LOGIN_DATA]?obj.loginData[LOGIN_DATA]:"");
        if(obj.loginData[LOGIN_ID]){
          if(post)post+="&";
          post+=obj.loginData[LOGIN_ID]+"="+encodeURIComponent(obj.user);
        }
        if(obj.loginData[LOGIN_PW]){
          if(post)post+="&";
          post+=obj.loginData[LOGIN_PW]+"="+encodeURIComponent(obj.password);
        }
        obj.loginData.splice(1,obj.loginData.length-1);
        obj.loginData[LOGIN_POST]=post;
      }
      obj.stage=obj.initStage;
      obj.inited=true;
    }catch(e){
dout(obj.id+" "+e);
      this.handlers.splice(i,1);//not loaded
      --i;
    }
  }
  var ar0=[];
  var ar1=[]; 
  for each(var o in this.handlers){
    if(o.order!=null){
      var ind=ar1.length;
      for(var i=0;i<ar1.length;i++){
        if(o.order<ar1[i].order){
          ind=i;
          break;
        }
      }      
      ar1.splice(ind,0,o);
    }
    else ar0.push(o);
  }
  for each(var o in ar1)ar0.splice(o.order,0,o);
  this.handlers=ar0;
  
  for(i=0;i<this.handlers.length;++i){  
    this.handlers[i].ind=i;
  }
  
  //if(!onInit)this.setState(nsIWebMailNotifier.ST_RESET,0);
  this.checkAll();
}
LightSMS.prototype.checkAll = function() {
  this.worker=null;
  this.setState(nsILightSMS.ST_LOGIN,nsILightSMS.RS_TRY);
  for each(var o in this.handlers){
    if(o.enabled)o.checkData();
  }
}
LightSMS.prototype.sortFunc = function(a,b) {
  var keyA=a.key?a.key:a.id;
  var keyB=b.key?b.key:b.id;
  if(keyA<keyB)return -1;
  else if(keyA==keyB){
    if(a.user<b.user)return -1;
    else if(a.user==b.user)return 0;
    else return 1;
  }else return 1;
}
LightSMS.prototype.stopAll = function() {
  //this.workTimer.cancel();
  for each(var o in this.handlers){
    if(o.started)o.stop();
  }
}
LightSMS.prototype.loadScript = function(id,isTemp) {
  var loader =  Components.classes["@mozilla.org/moz/jssubscript-loader;1"]
                  .getService(Components.interfaces.mozIJSSubScriptLoader);
  if(!this.scripts[id]){
//dout("loadScript "+(this.userScripts[id]?this.dir1:this.dir0)+id+".js");
    var scr={};
    loader.loadSubScript(this.dir0+id+".js",scr);
//dout("loaded "+id+" "+isTemp);
    if(!isTemp)this.scripts[id]=scr;
//if(!isTemp)dout("load ["+id+"] "+scr.name);
    return scr;
  }else return this.scripts[id];
}
LightSMS.prototype.getHostIDs = function(count) {
  var ar=[];
  for(var i in this.scripts){
    var o={id:i};
    if(this.order[i])o.key=this.order[i].toString();
    ar.push(o);
  }
  ar.sort(this.sortFunc);
  var ar2=[];
  for each(var o in ar)ar2.push(o.id);
  count.value=ar2.length;
  return ar2;
}
LightSMS.prototype.getHostName = function(id) {
  try{
    return this.bundle.GetStringFromName(id);
  }catch(e){}
  try{
    var obj=this.loadScript(id,true);
    return obj.name?obj.name:id;
  }catch(e){
dout("getHostName:"+id+" "+e);
    return id;
  };
}
LightSMS.prototype.getScriptVal = function(id,name) {
  try{
    var obj=this.loadScript(id,true);
    if(typeof obj[name]=="undefined")return null;
    return obj[name];
  }catch(e){
    return null;
  };
}
LightSMS.prototype.getAccountsNumber = function() {
  return this.handlers.length;
}
LightSMS.prototype.getAccountInfo = function(aIndex) {
  var obj=this.handlers[aIndex];
  var o={};
  o.id=obj.id;
  o.user=obj.user;
  o.password=obj.password;
  o.ind=obj.ind;
  o.count=obj.count;
  return encodeObject(o);
}
LightSMS.prototype.setState2 = function(o,error) {
  o.started=false;
  if(!this.worker){
    if(this.prefBranch.getBoolPref("useFaster")){
      this.setWorker();
    }else{
      var finished=true;
      for each(o in this.handlers){
        if(o.started){
          finished=false;
          break;
        }
      }
      if(finished){
        /*var total=0;
        for each(o in this.handlers){
          if(o.enabled&&o.count>0){
            total+=o.count;
          }
        }
  dout("all done "+total);*/
        this.setWorker();       
      }
    }
  }else if(o==this.worker){    
    if(o.count>=0)this.setCount(o);      
    if(o.count>0){
      if(this.msg.length>0){
        this.setState(nsILightSMS.ST_SEND,nsILightSMS.RS_TRY,this.getProgressStr());        
        o.send();
      }
    }else if(o.count==0){
      this.checkAll();
    }else{
      if(o.retry<1){
        ++o.retry;
        o.checkData();
      }else{
        this.checkAll();
      }
    }
  }
}
LightSMS.prototype.getWorkerStr=function(){
  return this.getHostName(this.worker.id)+" ["+this.worker.user+"]";
}
LightSMS.prototype.getProgressStr=function(){
  return this.msg[0][M_TOTAL]==1?"":" ["+(this.msg[0][M_INDEX]+1)+"/"+this.msg[0][M_TOTAL]+"]";
}
LightSMS.prototype.setCount=function(o){
  this.setState(nsILightSMS.ST_FREECOUNT,o.count,this.getWorkerStr());
}
LightSMS.prototype.setWorker=function(){
  var found=false;
  for each(o in this.handlers){
    if(!o.started&&o.enabled&&o.count>0){
      if(o==this.worker)continue;
      this.worker=o;
      found=true;
      break;
    }        
  }
  if(!found)this.worker=null;

  if(this.worker&&this.worker.count>=0){
    this.setState(nsILightSMS.ST_LOGIN,nsILightSMS.RS_SUCC);
    this.setCount(this.worker);
  }
  if(this.msg.length>0){
    if(this.worker==null){
      this.msg=[];
      this.setState(nsILightSMS.ST_LOGIN,nsILightSMS.RS_FAIL);
    }else{
      this.worker.checkData();
    }
  } 
}

LightSMS.prototype.setState = function(aID,aCount,aData) {
  for(var i=0;i<this.listeners.length;i++){
    if(this.listeners[i]){
      this.listeners[i].onStateChange(aID,aCount,aData);
    }
  }
}

/***********************************************************
preference observer
***********************************************************/
LightSMS.prototype.observe=function(aSubject, aTopic, aData)
{
  switch(aTopic){
  case "nsPref:changed":
    switch (aData) {
    case "showStatusbarIcon":
      this.setState(nsILightSMS.ST_ICON,this.prefBranch.getBoolPref("showStatusbarIcon"));
      break;
    }
    break;
  case "http-on-modify-request":
    for each(var i in this.handlers){
      if(i.channel==aSubject){
        i.onRequest(aSubject.QueryInterface(Components.interfaces.nsIHttpChannel));
        break;
      }
    }
    break;
  case "http-on-examine-response":
    for each(var i in this.handlers){
      if(i.channel==aSubject){
        i.onResponse(aSubject.QueryInterface(Components.interfaces.nsIHttpChannel));
        break;
      }
    }
    break;    
  }
}

/***********************************************************
password manager
***********************************************************/
LightSMS.prototype.getPassword=function(host,user){
  if(this.loginManager){
    var logins = this.loginManager.findLogins({},host,formSubmitURL, null);
    for each (var o in logins) {
      if (o.username==user) {
        return o.password;
      }
    }
  }else{
    var e = this.passwordManager.enumerator;
    while (e.hasMoreElements()) {
      try {
        var pass = e.getNext().QueryInterface(Components.interfaces.nsIPassword);
        if (pass.host == host&&pass.user == user) {
           return pass.password;
        }
      }catch(ex) {}
    }
  }
  return null;
}
LightSMS.prototype.setPassword=function(host,user,passwd){
  if(!passwd)return;
  if(this.loginManager){
    var nsLoginInfo = new Components.Constructor("@mozilla.org/login-manager/loginInfo;1",
                                                 Components.interfaces.nsILoginInfo,
                                                 "init");
    var loginInfo = new nsLoginInfo(host, formSubmitURL, null, user, passwd, "", "");

    var logins = this.loginManager.findLogins({},host,formSubmitURL, null);
    for each (var o in logins) {
      if (o.username==user) {
        this.loginManager.modifyLogin(o,loginInfo);
        return;
      }
    }
    this.loginManager.addLogin(loginInfo);
  }else{
    try{
      this.passwordManager.removeUser(host,user);
    }catch(e){}
    this.passwordManager.addUser(host,user,passwd);
  }
}
LightSMS.prototype.removePassword=function(host,user){
  if(this.loginManager){
    var logins = this.loginManager.findLogins({},host,formSubmitURL, null);
    for each (var o in logins) {
      if (o.username==user) {
        this.loginManager.removeLogin(o);
        break;
      }
    }
  }else{
    try{
      this.passwordManager.removeUser(host,user);
    }catch(e){}
  }
}


LightSMS.prototype.getNumber=function(str){
  return str.replace(/[-\(\)]/g,"").match(/\d{3,}/);
}
LightSMS.prototype.getDateString=function(){
      var dt=new Date();
      function getString(n){
        if(n<10)return "0"+n;
        else return n;
      }
      return dt.getFullYear()+"-"+getString(dt.getMonth()+1)+"-"+getString(dt.getDate())
              +" "+getString(dt.getHours())+":"+getString(dt.getMinutes())+":"+getString(dt.getSeconds());

}
LightSMS.prototype.onShowResult=function(succ){
  if(succ){
    this.setState(nsILightSMS.ST_SEND,nsILightSMS.RS_SUCC,this.getProgressStr());
  //  if(this.userData[DATA_FREE_COUNT]>0)--this.userData[DATA_FREE_COUNT];
  //  else if(this.userData[DATA_COUPON_COUNT]>0)--this.userData[DATA_COUPON_COUNT];  
    var m=this.msg.shift();
    if(this.prefBranch.getBoolPref("saveHist")){
      var str=m[M_MSG]+"\t"+m[M_RECV]+"\t"+m[M_SEND]+"\t"+this.getDateString();
      var db = Components.classes["@mozilla.org/autocomplete/search;1?name=text-file"]
                      .getService (Components.interfaces.nsITextFile);
      db.addEntry("smshist.txt", str,true);
    }
    if(this.msg.length>0){
      this.worker.checkData();
    }    
  }else{
    this.setState(nsILightSMS.ST_SEND,nsILightSMS.RS_FAIL,null);  
    if(this.msg.length>0)this.checkAll();
  }
}
LightSMS.prototype.getWindow=function(type){
  var wm = Components.classes["@mozilla.org/appshell/window-mediator;1"]
            .getService(Components.interfaces.nsIWindowMediator);
  var win = wm.getMostRecentWindow(type);
  if(win==null){
    win=Components.classes["@mozilla.org/appshell/appShellService;1"]
                    .getService(Components.interfaces.nsIAppShellService).hiddenDOMWindow;
  }
  return win;
}
LightSMS.prototype.openAuthDialog=function(){
  var win=this.getWindow("LightSMS:main");
  var params = {out:null};
  var features = "chrome,titlebar,toolbar,centerscreen,modal";
  win.openDialog("chrome://lightsms/content/auth.xul","",features,params).focus();
  if(params.out)return params.out.auth;
  else return null;  
}

/***********************************************************
class factory

This object is a member of the global-scope Components.classes.
It is keyed off of the contract ID.
***********************************************************/
var LightSMSFactory = {
  createInstance: function (aOuter, aIID)
  {
    if (aOuter != null)
      throw Components.results.NS_ERROR_NO_AGGREGATION;
    return (new LightSMS()).QueryInterface(aIID);
  }
};

/***********************************************************
module definition (xpcom registration)
***********************************************************/
var LightSMSModule = {
  registerSelf: function(aCompMgr, aFileSpec, aLocation, aType)
  {
    aCompMgr = aCompMgr.
        QueryInterface(Components.interfaces.nsIComponentRegistrar);
    aCompMgr.registerFactoryLocation(CLASS_ID, CLASS_NAME,
        CONTRACT_ID, aFileSpec, aLocation, aType);
  },

  unregisterSelf: function(aCompMgr, aLocation, aType)
  {
    aCompMgr = aCompMgr.
        QueryInterface(Components.interfaces.nsIComponentRegistrar);
    aCompMgr.unregisterFactoryLocation(CLASS_ID, aLocation);
  },

  getClassObject: function(aCompMgr, aCID, aIID)
  {
    if (!aIID.equals(Components.interfaces.nsIFactory))
      throw Components.results.NS_ERROR_NOT_IMPLEMENTED;

    if (aCID.equals(CLASS_ID))
      return LightSMSFactory;

    throw Components.results.NS_ERROR_NO_INTERFACE;
  },
  canUnload: function(aCompMgr) { return true; }
};

/***********************************************************
module initialization

When the application registers the component, this function
is called.
***********************************************************/
function NSGetModule(aCompMgr, aFileSpec) { return LightSMSModule; }




function Handler(main){
  this.main=main;
  this.cookieManager=new CookieManager();
  this.reset();
}
Handler.prototype={
  charSet: null,
  data: "",
  stage: 0,
  initStage: ST_LOGIN,
  count:-1,
  user: "",
  password: null,
  loginData:[],
  started:false,
  retry:0,
  // nsIChannelEventSink
  onChannelRedirect: function (aOldChannel, aNewChannel, aFlags) {
    if(this.channel==aOldChannel){
      this.channel = aNewChannel;
      if(this.defaultHeaders){
        try{
          var httpChannel = this.channel.QueryInterface(Components.interfaces.nsIHttpChannel);
          for(var t in this.defaultHeaders){
            httpChannel.setRequestHeader(t,this.defaultHeaders[t],false);
          }          
        }catch(e){
  dout(e);      
        }
        try{        
          aOldChannel = aOldChannel.QueryInterface(Components.interfaces.nsIHttpChannel);
          var ref=aOldChannel.getRequestHeader("Referer");
          if(ref)httpChannel.setRequestHeader("Referer",ref,false);
        }catch(e){}        
      }
    }
  },
  // nsIInterfaceRequestor
  getInterface: function (aIID) {
    try {
      return this.QueryInterface(aIID);
    } catch (e) {
//dout("getInterface "+aIID);
      throw Components.results.NS_NOINTERFACE;
    }
  },
  // nsIProgressEventSink
  onProgress : function (aRequest, aContext, aProgress, aProgressMax) { },
  onStatus : function (aRequest, aContext, aStatus, aStatusArg) { },
  // nsIHttpEventSink
  onRedirect : function (aOldChannel, aNewChannel) { },
  // nsIAuthPromptProvider
  getAuthPrompt : function(aReason){ return null; },
  QueryInterface : function(aIID) {
    if (aIID.equals(Components.interfaces.nsISupports) ||
        aIID.equals(Components.interfaces.nsIInterfaceRequestor) ||
        aIID.equals(Components.interfaces.nsIChannelEventSink) ||
        aIID.equals(Components.interfaces.nsIProgressEventSink) ||
        aIID.equals(Components.interfaces.nsIHttpEventSink) ||
        aIID.equals(Components.interfaces.nsIStreamListener)||
        aIID.equals(Components.interfaces.nsIAuthPromptProvider)||
        aIID.equals(Components.interfaces.nsIPrompt)|| //not used but required in ff 2.0
        aIID.equals(Components.interfaces.nsIDocShell)|| //required in ff 3.0
        aIID.equals(Components.interfaces.nsIAuthPrompt))
      return this;
//dout("QueryInterface "+aIID);
    throw Components.results.NS_NOINTERFACE;
  },
  onStartRequest : function (aRequest, aContext) {
    this.data = "";
  },
  onDataAvailable : function (aRequest, aContext, aStream, aSourceOffset, aLength){
	  var is = Components.classes["@mozilla.org/scriptableinputstream;1"]
            .createInstance(Components.interfaces.nsIScriptableInputStream);
  	is.init(aStream);
  	this.data += is.read(aLength);
  },
  onStopRequest : function (aRequest, aContext, aStatus) {
    if(aStatus==Components.results.NS_BINDING_ABORTED)return;
    var charSet=null;
    var httpChannel=aRequest.QueryInterface(Components.interfaces.nsIHttpChannel);
    if(this.charSet)charSet=this.charSet;
    else{
      var fnd=this.data.match(/<meta.+?[\"\'].+?charset=(\S+?)\s*?;?\s*?[\"\']\s*?\x2f?>/);
      if(fnd)charSet=fnd[1];
      else{
        try{
          if(httpChannel.contentCharset)charSet=httpChannel.contentCharset;
        }catch(e){}
      }
    }
    if(charSet){    
      var uniConv = Components.classes["@mozilla.org/intl/scriptableunicodeconverter"]
      								.createInstance(Components.interfaces.nsIScriptableUnicodeConverter);
      try{
        uniConv.charset = charSet;      
        this.data=uniConv.ConvertToUnicode(this.data);
      }catch(e){
      }
    }
    try{
      if(!this.process(aContext.QueryInterface(Components.interfaces.nsIHttpChannel), this.data))++this.stage;
    }catch(e){
      this.onError();
dout(this.id+" "+e);
    }
  },
  stop : function(){
    if(this.channel){
      this.channel.cancel(Components.results.NS_BINDING_ABORTED);
    }
  },
  process : function (aHttpChannel, aData){  
    switch(this.stage){
    case ST_LOGIN:
      this.getHtml(this.loginData[LOGIN_URL],this.loginData[LOGIN_POST]);
      return false;
    case ST_LOGIN_RES:
      this.stage=ST_DATA;
    case ST_DATA:
      this.getHtml(this.dataURL);
      return false;
    case ST_DATA_RES:
      var n=parseInt(this.getCount(aData));
      this.count=isNaN(n)?-1:n;
      if(this.count<0)this.reset();
      else{
        this.retry=0;
        this.stage=ST_DATA;
      }      
      this.main.setState2(this,false);
      return true;
    case ST_SEND:
      var m=this.main.msg[0];
      var m_kr=Components.classes["@mozilla.org/intl/texttosuburi;1"]
              .getService(Components.interfaces.nsITextToSubURI)
                .ConvertAndEscape("euc-kr",m[M_MSG]);
      return this.onSend(m[M_SEND],this.main.getNumber(m[M_RECV]),m_kr);
    case ST_SEND_RES:
      var fnd=aData.match(this.pattern);
      if(fnd)--this.count;      
      this.main.onShowResult(fnd);
      return true;      
    }
dout("[onError] "+this.id+" "+this.user+" "+this.stage);
    this.onError();
    return true;
  },
  onError : function(){
    this.reset();    
    this.main.setState2(this,true);      
  },
  reset : function(){
    this.count=-1;
    this.stage=this.initStage;
    this.cookieManager.clear();
  },
  checkData : function(){
    if(!this.enabled)return;
    this.stop();
    if(this.count<0)this.reset();
    else this.stage=ST_DATA;
    this.started=true;
    try{
      if(!this.process(null,""))++this.stage;
    }catch(e){
      this.onError();
dout(this.id+" "+e);
    }    
  },
  onRequest : function(aHttpChannel){
    var ck=this.cookieManager.getCookie(aHttpChannel.URI);
    aHttpChannel.setRequestHeader("Cookie",ck, false);
  },
  onResponse : function(aHttpChannel){
    try {
      var cookie = aHttpChannel.getResponseHeader("Set-Cookie");
      this.cookieManager.addCookies(aHttpChannel.URI,cookie);
      aHttpChannel.setResponseHeader("Set-Cookie", "", false);
    }catch(e){}
  },
  getIconURL : function(){
    return null;
  },
  getCount : function(aData){
    return -1;
  },
  send : function(){
    if(!this.enabled)return;
    this.stop();
    this.stage=ST_SEND;
    this.started=true;
    try{
      if(!this.process(null,""))++this.stage;
    }catch(e){
      this.onError();
dout(this.id+" "+e);
    }    
  },
  onSend : function onSend(sender,receiver,msg){
    return false;
  },
  setSender : function(mdn){
    var pref=this.main.prefBranch;
    var val=null;
    try{
      val=pref.getCharPref("sender");
    }catch(e){}
    if(!val)pref.setCharPref("sender",mdn);
  }
};
Handler.prototype.baseProcess=Handler.prototype.process
/***********************************************************
load html
***********************************************************/
Handler.prototype.getHtml = function(aURL,aPostData,aHeaders) {
  var ioService = Components.classes["@mozilla.org/network/io-service;1"].getService(Components.interfaces.nsIIOService);
  var uri = ioService.newURI(aURL, null, null);
  var channel = ioService.newChannelFromURI(uri);
  var httpChannel = channel.QueryInterface(Components.interfaces.nsIHttpChannel);
  //httpChannel.setRequestHeader("User-Agent","Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.0)",false);  
  if (aPostData||aPostData=="") {
    var uploadStream = Components.classes["@mozilla.org/io/string-input-stream;1"]
                        .createInstance(Components.interfaces.nsIStringInputStream);
    uploadStream.setData(aPostData, aPostData.length);
    var uploadChannel = channel.QueryInterface(Components.interfaces.nsIUploadChannel);
    uploadChannel.setUploadStream(uploadStream, "application/x-www-form-urlencoded", -1);
    httpChannel.requestMethod = "POST";
  }
  if(this.defaultHeaders){
    for(var t in this.defaultHeaders){
      httpChannel.setRequestHeader(t,this.defaultHeaders[t],false);
    }
  }    
  if(aHeaders){
    for(var t in aHeaders){
      httpChannel.setRequestHeader(t,aHeaders[t],false);
    }
  }
  this.channel=channel;
  channel.notificationCallbacks = this;
  channel.asyncOpen(this,httpChannel);
}

Handler.prototype.getForm=function(data,name,kr){
  if(name){
    var reg=new RegExp("<form.+?name\\s*=\\s*[\"\']"+name+"[\"\']([\\S\\s]+?)<\/form>","i");  
    var s=data.match(reg);  
    if(!s)return "";
    data=s[1];
  }
  var re=/<input[^>]+?name\s*=\s*[\"\'](\S+?)[\"\'][^>]+?value\s*=s*[\"\'](.*?)[\"\']/ig;
  var o;
  var post="";
  var conv=Components.classes["@mozilla.org/intl/texttosuburi;1"]
              .getService(Components.interfaces.nsITextToSubURI) ;               
  while ((o = re.exec(data)) != null){
    if(post)post+="&";
    post+=o[1]+"="+(kr?conv.ConvertAndEscape("euc-kr",o[2]):encodeURIComponent(o[2]));
  }
  return post;  
}

function encodeObject(data){
  var s="";
  for(var i in data){
    if(s)s+=",";
    s+=escape(i)+":"+escape(data[i]);
  }
  return s;
}
